# -*- coding: utf-8 -*-
import numpy as np
import pickle
import copy
from sklearn.metrics import roc_curve
from sklearn.metrics import auc
from itertools import combinations
from Bio.PDB import *
import matplotlib as mpl
mpl.use('pdf')
import matplotlib.pyplot as plt
import tensorflow as tf




def water_glue_model(L=18,K=4, n=1):
    '''
    '''
    import keras
    import neural_structured_learning as nsl
    from keras.layers.pooling import GlobalMaxPooling1D
    from keras.models import Sequential,Model
    from keras.layers import Dense,Activation,Dropout,Flatten
    from keras.layers.convolutional import Conv1D
    from keras.layers import Input, Embedding, Dense
    from keras.layers.normalization import BatchNormalization
    from keras.regularizers import l1,l2
    from keras.optimizers import SGD,RMSprop,Adam
    import tensorflow as tf
    #config = tf.ConfigProto(device_count={"CPU": 64})
    #keras.backend.tensorflow_backend.set_session(tf.Session(config=config))

    def residual_unit(nb_fil,f_len,ar,indx,residual='yes'):
        def f(input_node):
            bn1 = tf.keras.layers.BatchNormalization(name='BatchNormalization_'+indx+'_1')(input_node)
            act1 = tf.keras.layers.Activation('relu',name='relu_'+indx+'_1')(bn1)
            conv1= tf.keras.layers.Conv1D(nb_fil, f_len,dilation_rate=ar,kernel_regularizer=l2(0.0005), padding='same',name='conv1d_'+indx+'_1')(act1)
            bn2 = tf.keras.layers.BatchNormalization(name='BatchNormalization_'+indx+'_2')(conv1)
            act2 = tf.keras.layers.Activation('relu',name='relu_'+indx+'_2')(bn2)
            conv2= tf.keras.layers.Conv1D(nb_fil, f_len,dilation_rate=ar,kernel_regularizer=l2(0.0005),padding='same',name='conv1d_'+indx+'_2')(act2)
            if residual=='yes':
              output_node = tf.keras.layers.add([conv2, input_node])
            else:
              output_node = conv2
            return output_node
        return f
    num_vectors=50
    num_atoms = 30
    n_dim=3
    N=np.asarray([n,n])  #Depth of the model
    input0 = tf.keras.layers.Input(shape=(num_vectors,3), name='atom_vectors')
    input1 = tf.keras.layers.Input(shape=(num_vectors,num_atoms),name='beta')
    conv1 = tf.keras.layers.Conv1D(L, 1,kernel_regularizer=l2(0.0005), name='1d_conv1')(input0)
    conv2 = tf.keras.layers.Conv1D(L, 1,kernel_regularizer=l2(0.0005), name='1d_conv2')(input1)     
    merge1=tf.keras.layers.concatenate([conv1, conv2],name='merge12a')
    merge2=tf.keras.layers.concatenate([conv1, conv2],name='merge12b')
    conv=tf.keras.layers.Conv1D(L, K, name='1d_conv_reduce',kernel_regularizer=l2(0.0005),padding='same')(merge1) # see if 1 better or W
    skip=tf.keras.layers.Conv1D(L, K, name='1d_skip_reduce',padding='same')(merge2) # see if 1 better or W    
    for i in range(len(N)):
        for j in range(N[i]):
            conv = residual_unit(L, K, 1,'residual_'+str(i)+'_'+str(j),residual='yes')(conv)
        conv_for_skip = tf.keras.layers.Conv1D(L, 1,kernel_regularizer=l2(0.0005),name='convforskip'+str(i))(conv)    
        skip = tf.keras.layers.add([skip, conv_for_skip],name='skip'+str(i))
    conv = residual_unit(L, 1, 1,'residual_final',residual='yes')(skip)
    conv= tf.keras.layers.Conv1D(1, 1,padding='same',kernel_regularizer=l2(0.0005), activation='sigmoid',name='conv_final')(conv)
    output0=tf.keras.layers.GlobalMaxPooling1D(name='output')(conv)
    model=tf.keras.Model(inputs=[input0,input1],outputs=[output0])
    adv_config = nsl.configs.make_adv_reg_config(multiplier=0.2, adv_step_size=0.05)
    adv_model = nsl.keras.AdversarialRegularization(model, label_keys=['output'], adv_config=adv_config )
    adv_model.compile(optimizer='sgd',loss='binary_crossentropy',metrics=['accuracy'])
    return adv_model





# -*- coding: utf-8 -*-
from Bio.PDB import *
from atom_type import atom_type
import numpy as np
import sys
import os
import random
from itertools import combinations
from scipy.stats import mannwhitneyu
from scipy.stats import chi2_contingency
from numpy import linalg as LA
import warnings
import gzip
import shutil
import glob
import pickle
from multiprocessing import Process, JoinableQueue, Queue
import matplotlib as mpl
mpl.use('pdf')
import matplotlib.pyplot as plt
from collections import OrderedDict
from scipy.spatial import cKDTree

plt.rc('font', family='sans-serif', serif='Arial')
plt.rc('xtick', labelsize=16)
plt.rc('ytick', labelsize=14)
plt.rc('axes', labelsize=18)
plt.rc('axes', titlesize=18)
# we get all coordinates, b-factors, structure resolution and

AAs= set('ALA,ARG,ASN,ASP,CYS,GLN,GLU,GLY,HIS,ILE,LEU,LYS,MET,PHE,PRO,SER,THR,TRP,TYR,VAL'.split(','))
metal_res_names = ' ZN, CA,PO4, MG, MN, FE, CO, NA,  K'.split(',')
atom_names = [i.strip() for i in 'C   ,  CA  ,  CB  ,  CD  ,  CD1 ,  CD2 ,  CE  ,  CE1 ,  CE2 ,  CE3 ,  CG  ,  CG1 ,  CG2 ,  CH2 ,  CZ  ,  CZ2 ,  CZ3 ,  H   ,  HA  ,  HB  ,  HC  ,  HD1 ,  HD2 ,  HE  ,  HE1 ,  HE2 ,  HE3 ,  HG  ,  HG1 ,  HH  ,  HH2 ,  HO  ,  HOC ,  HZ  ,  HZ2 ,  HZ3 ,  N   ,  ND1 ,  ND2 ,  NE  ,  NE1 ,  NE2 ,  NH1 ,  NH2 ,  NZ  ,  O   ,  OC  ,  OD1 ,  OD2 ,  OE1 ,  OE2 ,  OG  ,  OG1 ,  OH  ,  SD  ,  SG  , 1H   , 1HA  , 1HB  , 1HD  , 1HD1 , 1HD2 , 1HE  , 1HE2 , 1HG  , 1HG1 , 1HG2 , 1HH1 , 1HH2 , 1HZ  , 2H   , 2HA  , 2HB  , 2HD  , 2HD1 , 2HD2 , 2HE  , 2HE2 , 2HG  , 2HG1 , 2HG2 , 2HH1 , 2HH2 , 2HZ  , 3HB  , 3HD1 , 3HD2 , 3HE  , 3HG1 , 3HG2 , 3HZ'.split(',')]
metals = 'ZN,CA,PO4,MG,MN,FE,CO,NA,K'.split(',')
valid_atoms = set(atom_names+metals +['P']) - {'PO4'}
pdbs = glob.glob('*.pdb')


def worker_task():
    while True:
        pdb = task_queue.get()
        if pdb is None:
            task_queue.task_done()
            break
        readPDB(pdb)
        task_queue.task_done()

def random_unit_vector():
    '''
    This generates a random unit vector
    '''
    generated=False
    while not generated:
        a = np.array([random.uniform(-1,1),random.uniform(-1,1),random.uniform(-1,1)])
        if sum(a*a) > 0 and sum(a*a) <= 1.0:
            generated=True
    return a/np.sqrt(sum(a*a))

def print_nonwaters_as_pdb(non_waters, waters, filename):
    #print ('number of non_waters is %s and number of waters is %s' % (len(non_waters), len(waters)))
    handle = open('%s_predicted_waters.pdb' % filename.split('.')[0], 'w')
    beginning =  'ATOM  61942  NA   NA C 989    '
    if non_waters:
        for non_water in non_waters:
            line = beginning+("%8.3f" % non_water[0])+("%8.3f" % non_water[1])+("%8.3f" % non_water[2]) +'  1.00  0.00'
            handle.write("%s\n" % line)
    beginning2 =  'HETATM%s O   HOH W%s'
    space = '    '
    if waters:
        for i in range(len(waters)):
            water=waters[i]
            line = (beginning2 % (100001+i, 1000+i))+space+("%8.3f" % water[0])+("%8.3f" % water[1])+("%8.3f" % water[2]) +'  1.00  0.00'
            handle.write("%s\n" % line)
    handle.close()


def cluster_predicted_waters(file_, threshold = 1.4):
    lines = open(file_).readlines()
    waters = [[float(i.strip().split()[-5]), float(i.strip().split()[-4]), float(i.strip().split()[-3])] for i in lines]
    waters = np.array(waters)
    added_waters = []
    handle = open('%s_clustered.pdb' % file_.split('.')[0], 'w')
    beginning =  'HETATM%s O   HOH W%s'
    space = '    '
    for i in range(len(waters)):
        water=waters[i]
        if len(added_waters) and min(LA.norm(added_waters-water, axis=1))<threshold:
            continue
        line = (beginning % (100001+i, 1000+i))+space+("%8.3f" % water[0])+("%8.3f" % water[1])+("%8.3f" % water[2]) +'  1.00  0.00'
        added_waters.append(water)
        handle.write("%s\n" % line)
    handle.close()


def read_surface(file_, atoms=None, print_inflated_mesh = False):
    '''
    This function reads in the coordinates defined by EDTSurf (run with option -f 20)
    '''
    handle = open(file_)
    scan_num_vertex = True
    scan_end_header = True
    counter=0
    coordinates = []
    for line in handle:
        if scan_num_vertex:
            if 'element vertex' in line:
                num_vertex = int(line.strip().split()[-1])
                scan_num_vertex = False
        if scan_end_header:
            if 'end_header' in line:
                counter=1
                scan_end_header=False
                continue
        if counter:
            parts = line.strip().split()
            assert('.' in parts[0])
            coordinates.append(map(float, parts[:3]))
            if counter==num_vertex:
                assert (len(coordinates)==num_vertex)
                coordinates = np.array(coordinates)
                handle.close()
                break
            counter+=1
    #if onion: legacy
    grid_sizes = []
    first_layer=-1.8
    first_layer2 = 0.0
    last_layer=5.8
    last_layer2=1.0
    for i in range(100):
        grid_sizes.append(sorted(LA.norm(coordinates- coordinates[random.choice(range(len(coordinates)))], axis=1))[1])
    grid_size = np.max(grid_sizes)
    if atoms is None:
        data, atoms = readPDB('%s.pdb' % (file_.split('.')[0]), pickle_=False, generate_nonwaters=False, return_atoms=True)
    else:
        data = {} # pass on empty data - this is when read surface is called from read_pdb
    atom_list = atoms.keys()
    atom_vectors = np.array(atom_list)
    proteins_centre = np.array([i.get_array() for i in atom_vectors]).mean(0)
    print ('Grid size is %s' % grid_size)
    num_layers_forward = int((last_layer-last_layer2)/grid_size) # we are seaching between 1.0 and 7.5A away
    num_layers_backward = int(-(first_layer-first_layer2)/grid_size)
    print ('Num onion layers is %s' % (num_layers_backward+num_layers_forward))
    print ('Num vertices is %s' % ((num_layers_backward+num_layers_forward)*num_vertex))
    coordinates_ = coordinates-proteins_centre # now the centre is at 0.0
    coordinates = [coordinates]
    mean_norm = np.mean(LA.norm(coordinates_, axis=1))
    scale_tick = grid_size/mean_norm
    print ('Scale tick size is %s' % scale_tick)
    start_forward = int(last_layer2/(mean_norm*scale_tick))
    start_backward = int(first_layer2/(mean_norm*scale_tick))
    coordinates = []
    for i in range(start_backward,num_layers_backward+start_backward):
        coordinates.append((1-(i)*scale_tick)*coordinates_+proteins_centre) # scale each layer and translate back to original position
    coordinates = coordinates[::-1]
    for i in range(start_forward, num_layers_forward+start_forward):
        coordinates.append((1+(i)*scale_tick)*coordinates_+proteins_centre) # scale each layer and translate back to original position
    for i in range(len(coordinates)):
        if print_inflated_mesh:
            handle = open(file_)
            handle2 = open("inflated_%s_%s.ply" % (file_.split('.')[0],i), "w")
            scan_num_vertex = True
            scan_end_header = True
            counter=0
            for line in handle:
                if scan_num_vertex:
                    if 'element vertex' in line:
                        num_vertex = int(line.strip().split()[-1])
                        scan_num_vertex = False
                if scan_end_header:
                    if 'end_header' in line:
                        handle2.write(line)
                        counter=1
                        scan_end_header=False
                        continue
                if counter:
                    handle2.write('%s %s\n' % (' '.join(map(str, coordinates[i][counter-1])),'255 255 255'))
                    counter+=1
                    if counter==num_vertex:
                        counter=False
                else:
                    handle2.write(line)
            handle.close()
            handle2.close()
    coordinates = np.array(coordinates)
    return coordinates, data, atoms

def parse_waters(atoms, waters, print_nonwaters=False, generate_nonwaters = True, nonwater_positions = None, all_nonwaters = False):
    # ensure that at least one atom is no firther than 5A from each water
    # we don't actually use the voxels
    '''
    print_nonwaters - for printing nonwater positions in PDB
    generate_nonwaters - flag specifying whether any (random or mesh) nonwaters are to be generated
    nonwater_positions - positions from the 'onion' mesh
    all_nonwaters - use all nonwaters from mesh, specifies the name for non-water pickled files
    Waters - key is original water position
    then we have vector from atom, atom type, presence in pocket voxel

    For each water, create random positions
    First find the geometric centre of the protein
    Then find atom on the protein's surface along a randomly generated vector
    Randomly choose a non-water position within 5A from this position and ensure that no real water is found closer than 2.5A from this position
    Run this function a desired number of times to get more negative data
    Alternatively, we can generate non-water positions by randomly selecting water molecules and choosing positions a random vector away (idea for the future)
    '''
    atom_list = atoms.keys() # list of atom vectors
    atom_vectors = np.array([i.get_array() for i in atom_list])
    non_waters = OrderedDict()
    waters_rejected=[] # if there are no atoms closer than 5A to the water, we reject that water molecule
    num_waters = len(waters)
    for water in waters:
        distances = atom_vectors-water.get_array()
        norms = LA.norm(atom_vectors-water.get_array(), axis=1)
        neighbor_norms = norms < 7.5
        neighbor_norms2 = norms < 4.6 # this just introduces a closer threshold to only include better examples
        if sum(neighbor_norms2)>3:
            counter_=0
            for i in neighbor_norms:
                if i:
                    if 'vectors' not in waters[water]:
                        waters[water]['vectors'] = [(Vector(distances[counter_]), atoms[atom_list[counter_]])]
                    else:
                        waters[water]['vectors'].append((Vector(distances[counter_]), atoms[atom_list[counter_]]))
                counter_+=1
        if len(waters[water])==0:
            waters_rejected.append(water)
    water_array = np.array([i.get_array() for i in waters.keys()])
    if generate_nonwaters:
        print ('generating non_waters')
        assert (nonwater_positions is not None)
        num_to_generate=len(waters)-len(waters_rejected)
        if all_nonwaters: # this is to use nonwaters from EDTSurf
            # check how many have already been generated previously
            already_generated = glob.glob('%s_[0-9].pickle' % (all_nonwaters))+glob.glob('%s_[0-9][0-9].pickle' % (all_nonwaters))
            if already_generated:
                already_generated = sorted([int(i.split('_')[-1].split('.')[0]) for i in already_generated])[-1]
                print("Already generated layers %s" % already_generated)
            else:
                already_generated=0
            num_to_generate = len(nonwater_positions)*len(nonwater_positions[0])
            total_to_generate = num_to_generate
            num_to_generate = num_to_generate-len(nonwater_positions[0])*already_generated
            print ("Number left to generate %s" % num_to_generate)
            counter = len(nonwater_positions[0])*already_generated
        very_far = False
        while num_to_generate>0:
            generated=False
            while not generated:
                if not all_nonwaters:
                    # generate the water away from the voxel
                    random1=-1
                    while random1 <0 or random1>=len(nonwater_positions):
                        random1 = np.random.normal(loc=len(nonwater_positions)/2.0, scale=np.sqrt(1.5*len(nonwater_positions)))
                    random1 = int(random1)
                    random2 = random.randint(0, len(nonwater_positions[0])-1)
                    nonwater_position = nonwater_positions[random1][random2]
                    # generate non_water and make sure water is not already presnet there...
                    if not very_far:
                        while min(LA.norm(water_array-nonwater_position, axis=1).tolist()) < 1.4 : # double check that water  is not already present here
                            random1=-1
                            while random1 <0 or random1>=len(nonwater_positions):
                                random1 = np.random.normal(loc=len(nonwater_positions)/3, scale=np.sqrt(1.5*len(nonwater_positions)))
                            random1 = int(random1)
                            random2 = random.randint(0,len(nonwater_positions[0])-1)
                            nonwater_position = nonwater_positions[random1][random2]
                    else:
                        while min(LA.norm(water_array-nonwater_position, axis=1).tolist()) < 1.4 or min(LA.norm(atom_vectors-nonwater_position, axis=1).tolist()) >= 7.5: # double check that water  is not already present here
                            random1=-1
                            while random1 <0 or random1>=len(nonwater_positions):
                                random1 = np.random.normal(loc=len(nonwater_positions)/3, scale=np.sqrt(1.5*len(nonwater_positions)))
                            random1 = int(random1)
                            random2 = random.randint(0,len(nonwater_positions[0])-1)
                            nonwater_position = nonwater_positions[random1][random2]
                    nonwater_position_ = Vector(nonwater_position)
                    non_waters[nonwater_position_] = {}
                else:
                	#add code that checks that non-waters are already generates so that code continues later... (fix the 4uis & 1g6v runs)
                    nonwater_position = nonwater_positions[counter/len(nonwater_positions[0])][counter % len(nonwater_positions[0])] # the purpose of this is to add all nonwater-positions generated by EDTSurf
                    counter+=1
                    nonwater_position_ = Vector(nonwater_position)
                    non_waters[nonwater_position_] = {}
                distances = atom_vectors-nonwater_position
                norms = LA.norm(atom_vectors-nonwater_position, axis=1)
                neighbor_norms = norms < 7.5
                counter_=0
                for i in neighbor_norms:
                    if i or ((all_nonwaters or not very_far) and not np.max(neighbor_norms)): # the purpose of this or is to add at least one vector for nonwater even if too far - when dealing with EDTSurf mesh
                        if 'vectors' not in non_waters[nonwater_position_]:
                            if not i: # this is when all are greater (or equal) 7.5
                                non_waters[nonwater_position_]['vectors'] = [(Vector(distances[norms.tolist().index(np.min(norms))]), atoms[atom_list[norms.tolist().index(np.min(norms))]])]
                                very_far=True
                            else:
                                non_waters[nonwater_position_]['vectors'] = [(Vector(distances[counter_]), atoms[atom_list[counter_]])]
                        else:
                            non_waters[nonwater_position_]['vectors'].append((Vector(distances[counter_]), atoms[atom_list[counter_]]))
                    counter_+=1
                    if very_far and 'vectors' in non_waters[nonwater_position_] and len(non_waters[nonwater_position_]['vectors'])==50:
                        break
                if nonwater_position_ in non_waters:
                    generated=True
                    num_to_generate-=1
                    if not all_nonwaters:
                        print ('non_water generated, no. interacting atoms is %s' % len(non_waters[nonwater_position_]['vectors']))
                    if all_nonwaters and (total_to_generate-num_to_generate) % len(nonwater_positions[0])==0:
                        data = {'waters':waters, 'non_waters':non_waters}
                        pickle.dump(data, open('%s_%s.pickle' % (all_nonwaters, (total_to_generate-num_to_generate)/len(nonwater_positions[0])), 'wb'), protocol=pickle.HIGHEST_PROTOCOL)
                        non_waters = OrderedDict() # after dumping file I empty non_water dict in order to save memory
    print ("%s out of %s waters rejected" % (len(waters_rejected),len(waters)))
    for water in waters_rejected:
        del waters[water]
    if print_nonwaters:
        print_nonwaters_as_pdb(non_waters,waters, print_nonwaters)
    else:
        print ('print nonwaters set to false')
    return waters, non_waters, len(waters_rejected), len(waters_rejected)+num_waters


def readPDB(filename, print_nonwaters=False, pickle_=True, generate_nonwaters=True, return_atoms = False):
    """
    Returns atom coordinates & water data
    """
    data = {}
    handle = open(filename)
    resolution=0
    r_factor=0
    solvent_content =  0
    for line in handle:
        parts = line.strip().split()
        if len(parts) >=4  and parts[0]=='REMARK' and parts[1]=='2' and parts[2]=='RESOLUTION.':
            try:
                resolution=float(parts[3])
            except:
                continue
        elif len(parts) >=6  and parts[0]=='REMARK' and parts[1]=='3' and parts[2]=='R' and parts[3]=='VALUE' and parts[4]=='(WORKING' and 'SET' in parts[5]:
            try:
                r_factor = float(parts[-1])
            except:
                continue
        elif len(parts) >=5  and parts[0]=='REMARK' and parts[1]=='280' and parts[2]=='SOLVENT' and parts[3]=='CONTENT,' and parts[4]=='VS':
            try:
                solvent_content = float(parts[-1])
            except:
                continue
    handle.close()
    print('Reading %s PDB file, resolution: %s, R-factor: %s, solvent per-cent: %s' % (filename.split('/')[-1].split('.')[0],resolution, r_factor, solvent_content) )
    handle.close()
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        atoms = {}
        waters = {}
        p=PDBParser()
        structure=p.get_structure('name', filename)
        for model in structure:
            for chain in model:
                for residue in chain:
                    if residue.resname in AAs or residue.resname in metals:
                        for atom in residue.child_dict.keys():
                            if atom in valid_atoms:
                                if residue.resname in AAs:
                                    atoms[residue[atom].get_vector()] = (atom,residue.resname, residue[atom].get_bfactor())
                                elif residue.resname in metals:
                                    atoms[residue[atom].get_vector()] = (atom,None, residue[atom].get_bfactor())
                    elif residue.resname=='HOH': # TODO: add reading of metal ions
                        if residue['O'].get_occupancy()==1.0:
                            waters[residue['O'].get_vector()] = {}
        if generate_nonwaters:
            try:
                coordinates, data, atoms = read_surface("%s.ply" % filename.split('.')[0], atoms=atoms)
            except:
                os.system('./EDTSurf -i %s -o %s -f 20 -h 2' % (filename, filename.split('.')[0]))
                coordinates, data, atoms = read_surface("%s.ply" % filename.split('.')[0], atoms=atoms)
            waters, non_waters, waters_rejected, total_waters = parse_waters(atoms, waters, print_nonwaters=print_nonwaters, generate_nonwaters=generate_nonwaters,nonwater_positions =coordinates)
        else:
            waters, non_waters, waters_rejected, total_waters = parse_waters(atoms, waters, print_nonwaters=print_nonwaters, generate_nonwaters=generate_nonwaters)
        data = {'waters':waters, 'non_waters':non_waters, 'waters_rejected':waters_rejected, 'total_waters':total_waters, 'resolution': resolution, 'r_factor':r_factor, 'solvent_content':solvent_content}
        if pickle_:
            pickle.dump(data, open('%s.pickle' % filename.split('.')[0], "wb"), protocol=pickle.HIGHEST_PROTOCOL)
        if return_atoms:
            return data, atoms
        else:
            return data



def process_waters(waters, vectors, b_values, num_atoms_to_consider=50,output_original_coordinates= False):
    '''
    This runs preprocessing on (non)waters and reurns data in format usable as input in the ML model
    '''
    if output_original_coordinates:
        original_cooridinates = []
    for water in waters:
        if output_original_coordinates:
            original_cooridinates.append(water)
        temp = np.zeros([num_atoms_to_consider,3])
        temp2 = np.ones([num_atoms_to_consider,30]) # since 1 is a 'bad' beta factor
        water_vectors = [i[0].__pow__(1.0/7.5) for i in waters[water]['vectors']] # normalized such that they are between 0 and 1
        water_vector_norms = [i.norm() for i in water_vectors]
        if np.max(water_vector_norms) >1.0:
            vectors.append(temp)
            b_values.append(np.zeros([num_atoms_to_consider,30]))
            continue
        closest_ids = sorted(range(len(water_vector_norms)), key=lambda k: water_vector_norms[k])
        atoms = [i[1][0] for i in waters[water]['vectors']]
        resids = [i[1][1] for i in waters[water]['vectors']]
        #betas = [np.log10(1+i[1][2])/2.7 for i in waters[water]['vectors']] # this was the original normalization
        betas = [np.log10(1+i[1][2]) for i in waters[water]['vectors']]
        if np.max(betas)==np.min(betas):
            betas = [0]*len(betas)
        else:
            betas = [(i-np.min(betas))/(np.max(betas)-np.min(betas)) for i in betas]
        ### dtete this betas = [np.log10(1+i[1][2])/max([np.log10(1+j[1][2]) for j in waters[water]['vectors']]) for i in waters[water]['vectors']] # per protein normalization
        angles = np.zeros([len(water_vectors), len(water_vectors)])
        if len(water_vectors) >1:
            for comb in combinations(range(len(water_vectors)),2):
                angles[comb[0], comb[1]] = water_vectors[comb[0]].angle(water_vectors[comb[1]])
                angles[comb[1], comb[0]] = water_vectors[comb[0]].angle(water_vectors[comb[1]])
            e1 = water_vectors[closest_ids[0]].normalized() # closest atom, then largest angle to closest
            id_of_partner = list(angles[closest_ids[0]]).index(max(angles[closest_ids[0]]))
            e2 = water_vectors[id_of_partner]-e1.__pow__(e1.__mul__(water_vectors[id_of_partner]))
            e2.normalize()
            e3= e1.__pow__(e2)
            m = np.array([e3[:], e2[:], e1[:]]) # normalization rotation-matrix
            done_waters= set([closest_ids[0], id_of_partner])
            index = atom_type(atoms[id_of_partner], resids[id_of_partner])#atom_names.index(atoms[id_of_partner])# this is correct
            normalized_coordinates = water_vectors[id_of_partner].left_multiply(m)[:]
            temp2[1,index] = betas[id_of_partner] # append beta factor
            temp[1,:] = normalized_coordinates
            if len(done_waters) < len(water_vectors):
                for id_ in closest_ids[1:]:
                    if len(done_waters) == num_atoms_to_consider:
                        break
                    if id_ in done_waters:
                        continue
                    else:
                        done_waters.add(id_)
                        index = atom_type(atoms[id_], resids[id_])#atom_names.index(atoms[id_])# this is correct
                        normalized_coordinates = water_vectors[id_].left_multiply(m)[:] # normalize coordinates
                        temp2[len(done_waters)-1,index] = betas[id_] # append beta factor
                        temp[len(done_waters)-1,:] = normalized_coordinates
                        if len(done_waters) < len(water_vectors) and len(done_waters) < num_atoms_to_consider:
                            ids_of_partners = sorted(range(len(angles[id_])), key=lambda k: angles[id_][k])
                            for id__ in ids_of_partners:
                                if id__ in done_waters:
                                    id_of_partner2 = id__
                                    break
                            done_waters.add(id_of_partner2)
                            index = atom_type(atoms[id_of_partner2], resids[id_of_partner2])#atom_names.index(atoms[id_of_partner2])# this is correct
                            normalized_coordinates = water_vectors[id_of_partner2].left_multiply(m)[:] # normalize coordinates
                            temp2[len(done_waters)-1,index] = betas[id_of_partner2] # append beta factor
                            temp[len(done_waters)-1,:] = normalized_coordinates
        else:
            e1 = water_vectors[closest_ids[0]].normalized()
            partner = Vector(3*random_unit_vector())
            e2 = partner-e1.__pow__(e1.__mul__(partner))
            e2.normalize()
            e3= e1.__pow__(e2)
            m = np.array([e3[:], e2[:], e1[:]]) # normalization rotation-matrix
        #at the end add first atom regardless whether we have one atom or more...
        index = atom_type(atoms[closest_ids[0]], resids[closest_ids[0]])#atom_names.index(atoms[closest_ids[0]])# this is correct
        normalized_coordinates = water_vectors[closest_ids[0]].left_multiply(m)[:] # normalize coordinates
        temp2[0,index] = betas[closest_ids[0]] # append beta factor
        temp[0,:] = normalized_coordinates
        vectors.append(temp)
        b_values.append(np.ones([num_atoms_to_consider,30]) -temp2)
    if output_original_coordinates:
        return vectors, b_values, original_cooridinates
    else:
        return vectors, b_values















import re

AAs= set('ALA,ARG,ASN,ASP,CYS,GLN,GLU,GLY,HIS,ILE,LEU,LYS,MET,PHE,PRO,SER,THR,TRP,TYR,VAL'.split(','))
atom_names = [i.strip() for i in 'C   ,  CA  ,  CB  ,  CD  ,  CD1 ,  CD2 ,  CE  ,  CE1 ,  CE2 ,  CE3 ,  CG  ,  CG1 ,  CG2 ,  CH2 ,  CZ  ,  CZ2 ,  CZ3 ,  H   ,  HA  ,  HB  ,  HC  ,  HD1 ,  HD2 ,  HE  ,  HE1 ,  HE2 ,  HE3 ,  HG  ,  HG1 ,  HH  ,  HH2 ,  HO  ,  HOC ,  HZ  ,  HZ2 ,  HZ3 ,  N   ,  ND1 ,  ND2 ,  NE  ,  NE1 ,  NE2 ,  NH1 ,  NH2 ,  NZ  ,  O   ,  OC  ,  OD1 ,  OD2 ,  OE1 ,  OE2 ,  OG  ,  OG1 ,  OH  ,  SD  ,  SG  , 1H   , 1HA  , 1HB  , 1HD  , 1HD1 , 1HD2 , 1HE  , 1HE2 , 1HG  , 1HG1 , 1HG2 , 1HH1 , 1HH2 , 1HZ  , 2H   , 2HA  , 2HB  , 2HD  , 2HD1 , 2HD2 , 2HE  , 2HE2 , 2HG  , 2HG1 , 2HG2 , 2HH1 , 2HH2 , 2HZ  , 3HB  , 3HD1 , 3HD2 , 3HE  , 3HG1 , 3HG2 , 3HZ'.split(',')]
metals = 'ZN,CA,P,MG,MN,FE,CO,NA,K'.split(',')

def atom_type(atom,resid=None):
    '''
    //simplest representation
    //1-aliphatic carbon and an alpha carbon- maybe separate later
    //2 - aromatic carbon, 3-amide nitrogen (main chain), 
    //4 - carbonyl carbon (main chain),
    //5- hydroxyl oxygen, 6-carboxyl oxygen, 7-hydroxyl aromatic, 8 - carbonyl oxygen(side chain), 
    //9 - amide nitrogen (side chain), 10-amide carbon (main chain), 11 - amide carbon (side chain)
    //12 - carboxyl carbon, 13-aromatic nitrogen (TRP), 14 - MET sulfur, 15 - CYS sulfur, 16-guanidinium carbon
    //17- aromatic nitrogen (HIS) 18-amine nitrogen (secondary), 19- amine nitrogen (primary), 20 - carbonyl oxygen(main chain)
    //21-29: ZN,CA,P,MG,MN,FE,CO,NA,K (metals)
    '''
    atom = re.match("[A-Z]+", atom).group() # remove numbers from atom definitions
    if resid is not None:
        if atom in set(['CA', 'CB']):
            return 0
        elif atom=='C':
            return 3
        elif atom=='CG':
            if resid in set(['GLN','GLU','ILE','LEU','LYS','MET','PRO','THR','VAL','ARG']):
                return 0
            elif resid in set(['PHE', 'TYR', 'TRP', 'HIS']):
                return 1
            elif resid=='ASN':
                return 10
            elif resid =='ASP':
                return 11
        elif atom =='CD':
            if resid in set(['ARG',"ILE","LEU","LYS","PRO"]):
                return 0
            elif resid=='GLN':
                return 10
            elif resid=='GLU':
                return 11
            elif resid in set(['PHE', 'TYR', 'TRP','HIS']):
                return 1
        elif atom=='CE':
            if resid in set(['HIS', 'PHE', 'TRP', 'TYR']):
                return 1
            elif resid in set(['LYS', 'MET']):
                return 0
        elif atom=='CZ':
            if resid in set(['PHE', 'TRP', 'TYR']):
                return 1
            elif resid=='ARG':
                return 15
        elif atom =='CH':
            return 1
        elif atom=='SG':
            return 14
        elif atom=='SD':
            return 13
        elif atom=='O':
            return 19
        elif atom=='OD':
            if resid=='ASN':
                return 7
            elif resid=='ASP':
                return 5
        elif atom=='OG':
            return 4
        elif atom=='OE':
            if resid=='GLN':
                return 7
            elif resid =='GLU':
                return 5
        elif atom=='OH':
            return 6
        elif atom=='N':
            return 2
        elif atom =='NE':
            if resid=='ARG':
                return 17
            elif resid=='HIS':
                return 16
            elif resid=='TRP':
                return 12
            elif resid =='GLN':
                return 8
        elif atom=='NH':
            return 18
        elif atom=='ND':
            if resid=='ASN':
                return 8
            elif resid=='HIS':
                return 16
        elif atom=='NZ':
            return 18
        else:
            print ("unidentified atom type" % atom)
            return 29
    else:
        return metals.index(atom) +20



'''
@author: Jan Zaucha. Created on 23 Jan 2020

Purpose: To predict water hot spots on the surface of proteins

MAIN methods here 
'''

import sys
from Bio.PDB import *
from h2o_bind import *
import numpy as np
import random
import time
import matplotlib as mpl
mpl.use('pdf')
import matplotlib.pyplot as plt
from numpy import linalg as LA
from model import *
import operator
import requests
path = '/'.join(os.path.realpath(__file__).split('/')[:-1])+'/'


def worker_task_ply():
    while True:
        print('trying to get data to process')
        to_process = task_queue_ply.get()
        print ('got data to process')
        if to_process is None:
            task_queue_ply.task_done()
            print ('breaking')
            break
        print("processing %s,%s,%s" % (to_process['file_'],to_process['model'],to_process['i']))
        process_layer(to_process['file_'],to_process['model'],to_process['i'])
        task_queue_ply.task_done()



def process_layers(file_, model, k,threads=4):
    global task_queue_ply
    task_queue_ply = JoinableQueue()
    for i in range(threads):
        print("starting thread %s" % i)
        p = Process(target=worker_task_ply)
        p.start()

    for j in range(k):
        task_queue_ply.put({'file_':file_,'model':model,'i':j})

    #Add a sentinal task for each child so that they know when to terminate
    for thread in range(threads):
        task_queue_ply.put(None)

    task_queue_ply.join()
    task_queue_ply.close()
    return True

def process_layer(file_,model, i):
    print ("processing layer %s" % i)
    vectors =[]
    b_values = []
    if os.path.isfile("%s_%s_processed.pickle" % (file_.split('.')[0], i+1)):
        print ("%s_%s_processed.pickle is present" % (file_.split('.')[0], i+1) )
        return True
    data = pickle.load(open("%s_%s.pickle" % (file_.split('.')[0], i+1), 'rb'))
    vectors, b_values ,original_coords= process_waters(data['non_waters'], vectors, b_values, output_original_coordinates= True)
    #y_pred = model['model'].predict({'atom_vectors':np.array(vectors) ,'beta':np.array(b_values)}).ravel().tolist()
    data_processed = {'vectors':vectors, 'b_values':b_values ,'original_coords':original_coords}
    pickle.dump(data_processed, open("%s_%s_processed.pickle" % (file_.split('.')[0], i+1), 'wb'),protocol=pickle.HIGHEST_PROTOCOL)


def define_vertices_for_waters(coordinates, real_waters, name, print_angles=True):
    '''
    real waters must be an array of vectors
    This identifies for all waters the closes vertices
    mesh layer: proportion of total waters in each onion layer
    '''

    occupied_by_waters = set()
    water_coordinates = []
    captured = []
    coord_trie =  cKDTree(coordinates[-1,:,:])
    grid_sizes = []
    for i in range(100):
        grid_sizes.append(coord_trie.query(coordinates[-1,random.choice(range(len(coordinates))),:], k=2)[0][1])
    grid_size = np.mean(grid_sizes)
    print ("Maximal grid size is about %s" % grid_size)
    for real_water in real_waters:
        dist2mesh = []
        index_min = []
        original_coord = []
        for j in range(len(coordinates)):
            distances = coordinates[j]-real_water.get_array() 
            norms = LA.norm(distances, axis=1).tolist()
            assert (len(norms)==len(coordinates[j]))
            index_min.append(norms.index(min(norms)))
            dist2mesh.append(min(norms))
            original_coord.append(coordinates[0][norms.index(min(norms))])
        if min(dist2mesh) < grid_size*np.sqrt(3):
            captured.append(True)
            occupied_by_waters.add(index_min[dist2mesh.index(min(dist2mesh))])
            #print("water captured %s, distance from mesh %s, index %s, original_coord %s" % (real_water.get_array(),min(dist2mesh), index_min[dist2mesh.index(min(dist2mesh))], original_coord[dist2mesh.index(min(dist2mesh))] ))
        else:
            #print("water missed %s, distance from mesh %s" % (real_water.get_array(),min(dist2mesh) ) )
            captured.append(False)
    if captured:
        frac_captured = sum(captured)/float(len(captured))
    else:
        frac_captured=1.0
    return occupied_by_waters, frac_captured, captured



def download_pdb_file(pdbcode):
    if os.path.isfile('%s.pdb' % pdbcode): 
        print ("Creating mesh around protein surface using EDTSurf")
        os.system('%s/EDTSurf -i %s.pdb -o %s -f 20 -h 2' % (path, pdbcode, pdbcode))
    else:
        try:
            print("Downloading PDB file from RCSB PDB: accession no. %s" % pdbcode)
            url = "http://files.rcsb.org/download/%s.pdb.gz" % pdbcode
            fname = '%s.pdb' % pdbcode
            r = requests.get(url)
            num_bytes = open('%spdb.gz' % pdbcode , 'wb').write(r.content)
            with gzip.open('%spdb.gz' % pdbcode, 'rb') as f_in:
                with open('%s.pdb' % pdbcode, 'wb') as f_out:
                    shutil.copyfileobj(f_in, f_out)
            print ("Creating mesh around protein surface using EDTSurf")
            os.system('%s/EDTSurf -i %s.pdb -o %s -f 20 -h 2' % (path,pdbcode, pdbcode))
        except:
            print("Error: Downloading/unpacking %s failed; skipping..." % pdbcode)
            sys.exit(0)


def scan_surface(pdbcode, threads=4, threshold=0.999):
    '''
    This fills the file white
    '''
    model_=path+'water_hot_spots_adv_new.h5'
    download_pdb_file(pdbcode)
    model = water_glue_model(L=30,K=4, n=1).base_model
    model.load_weights(model_)
    mesh, data, atoms = read_surface("%s.ply" % pdbcode)
    data['waters'] = OrderedDict(data['waters'])
    vertices_occupied, frac_captured, captured = define_vertices_for_waters(mesh, data['waters'],pdbcode)
    waters, non_waters, waters_rejected, total_waters = parse_waters(atoms, data['waters'], print_nonwaters=False, generate_nonwaters = True, nonwater_positions = mesh, all_nonwaters = pdbcode)
    predictions = np.zeros(np.array(mesh).shape[:2])
    process_layers("%s.ply" % pdbcode, model_, np.array(mesh).shape[0],threads=4)
    if not os.path.isfile("%s_pred_%s.pickle" % (pdbcode,threshold)):
        for i in range(0, np.array(mesh).shape[0]):
            print ("Trying command pickle load open %s_%s_processed.pickle" % (pdbcode, i+1))
            data_processed = pickle.load(open("%s_%s_processed.pickle" % (pdbcode, i+1), 'rb'))
            predictions[i,:] = model.predict({'atom_vectors':np.array(data_processed['vectors']) ,'beta':np.array(data_processed['b_values'])}).ravel().tolist()
        data2 = {'predictions':predictions, 'mesh':mesh}
        pickle.dump(data2, open("%s_pred_%s.pickle" % (pdbcode,threshold), 'wb'),protocol=pickle.HIGHEST_PROTOCOL)
    else:
        data2 = pickle.load(open("%s_pred_%s.pickle" % (pdbcode,threshold), 'rb'))
    collapsed_predictions = data2['predictions'].max(axis=0)
    print("Finished loading precalculated predictions")
    # get positions predicted as water (only maximum onion layer for each mesh point above chosen threshold)
    max_pred = np.amax(data2['predictions'],axis=0) # this is the maximum prediction at each point on surface
    arg_max =data2['predictions'].argmax(axis=0) # this is the index of the maximum points
    meets_threshold =max_pred>threshold
    index_if_meets_threshold = arg_max[meets_threshold]
    predicted_waters = []
    for i in range(data2['mesh'][:,meets_threshold].shape[1]):
        predicted_waters.append(data2['mesh'][:,meets_threshold][index_if_meets_threshold[i]][i])
    print_nonwaters_as_pdb(None, predicted_waters, pdbcode)
    print ('Finished printing predicted waters to pdb')
    os.system('rm %s-cav.pdb' % pdbcode)


def rank_waters(pdbcode):
    model_=path+'water_hot_spots_adv_new.h5'
    download_pdb_file(pdbcode)
    model = water_glue_model(L=30,K=4, n=1).base_model
    model.load_weights(model_)
    mesh, data, atoms = read_surface("%s.ply" % pdbcode)
    waters, non_waters, waters_rejected, total_waters = parse_waters(atoms, data['waters'], print_nonwaters=False, generate_nonwaters = False)
    vectors =[]
    b_values = []
    vectors, b_values ,original_coords= process_waters(data['waters'], vectors, b_values, output_original_coordinates= True)
    predictions = model.predict({'atom_vectors':np.array(vectors) ,'beta':np.array(b_values)}).ravel().tolist()
    results  =  {}
    for i,j in zip(predictions,original_coords):
        results[j]=i
    results = sorted(results.items(), key=operator.itemgetter(1))[::-1]
    handle = open('%s_water_ranking.txt' % pdbcode, 'w')
    handle.write('Water-position\tprediction-score\n')
    for i in results:
        handle.write('%s\t%s\n' % (i[0].get_array(),i[1]))
    handle.close()
    print ("Ranked water positions along with scores have been written out to a file")
    os.system('rm %s-cav.pdb' % pdbcode)




from setuptools import setup

def readme():
    with open('README.rst') as f:
        return f.read()

setup(name='hotWater',
      version='1.8',
      description='The program uses a residual deep learning model to predict water hot-spots on the surface of proteins',
      long_description=readme(),
      long_description_content_type='text/x-rst',
      author='Jan Zaucha',
      author_email='trelek2@gmail.com',
      license='MIT',
      packages=['hotWater'],
      include_package_data=True,
      install_requires=['numpy', 'scipy', 'biopython', 'matplotlib', 'scikit-learn', 'tensorflow', 'tensorflow-gpu', 'keras', 'neural_structured_learning'],
      dependency_links=['http://biopython.org/DIST/'],
      zip_safe=False)
